@@ -9,7 +9,10 @@ import hashlib
 import codecs
 import struct
 from cryptography.hazmat import backends
+from cryptography.hazmat.primitives import serialization
 from cryptography.hazmat.primitives.serialization import pkcs12
+from cryptography.hazmat.primitives.asymmetric import ec
+from cryptography.x509 import ObjectIdentifier
 from endesive import signer
 from endesive.pdf.PyPDF2 import pdf, generic as po
 
@@ -204,6 +207,7 @@ class SignedData(pdf.PdfFileWriter):
     def _make_signature(self, Contents=None, Type=None, SubFilter=None):
         sig = po.DictionaryObject()
         sig_ref = self._addObject(sig)
+
         sig.update(
             {
                 po.NameObject("/Type"): Type,
@@ -248,7 +252,165 @@ class SignedData(pdf.PdfFileWriter):
         )
         return annot, annot_ref
 
-    def makepdf(self, prev, udct, algomd, zeros, cert, **params):
+    def addAnnotation(self, cert, udct, box, page0ref, obj13, obj13ref, new_13):
+        from endesive.pdf.PyPDF2_annotate.annotations.signature import Signature
+        from endesive.pdf.PyPDF2_annotate.config.appearance import Appearance
+        from endesive.pdf.PyPDF2_annotate.config.location import Location
+        from endesive.pdf.PyPDF2_annotate.util.geometry import identity
+
+        x1, y1, x2, y2 = box
+        annotation = Signature(
+            Location(x1=x1, y1=y1, x2=x2, y2=y2, page=0), Appearance()
+        )
+        if "signature" in udct:
+            # Plain text signature with the default font
+            # text to render is contained in udct['signature']
+            # font parameters are in udct['signature']['text']
+            annotationtext = udct["signature"]
+            wrap_text = udct.get("text", {}).get("wraptext", True)
+            font_size = udct.get("text", {}).get("fontsize", 12)
+            text_align = udct.get("text", {}).get("textalign", "left")
+            line_spacing = udct.get("text", {}).get("linespacing", 1.2)
+
+            annotation.add_default_font()
+            annotation.set_signature_appearance(
+                ["fill_colour", 0, 0, 0],
+                ["font", "default", font_size],
+                [
+                    "text_box",
+                    annotationtext,
+                    "default",
+                    0,
+                    0,
+                    x2 - x1,
+                    y2 - y1,
+                    font_size,
+                    wrap_text,
+                    text_align,
+                    "middle",
+                    line_spacing,
+                ],
+            )
+        elif "signature_img" in udct:
+            # Simple image signature, stretches to fit the box
+            # image to render is contained in udct['signature_image']
+            annotation.add_image(udct["signature_img"], "Image")
+            annotation.set_signature_appearance(
+                [
+                    "image",
+                    "Image",
+                    0,
+                    0,
+                    x2 - x1,
+                    y2 - y1,
+                    udct.get("signature_img_distort", True),
+                    udct.get("signature_img_centred", False),
+                ]
+            )
+        elif "signature_appearance" in udct:
+            # Adobe-inspired signature with text and images
+            # Parameters are contained in udct['signature_appearance']
+            # If a field is included in the display list, that field
+            # will be contained in the annotation.
+            #
+            # Text and border are the colour specified by outline,
+            # and border is the the inset distance from the outer
+            # edge of the annotation.  The R G B values range between
+            # 0 and 1.
+            #
+            # Icon is an image to display above the background and
+            # border at the left-hand side of the anntoation.  If
+            # there is no text, it is centred.
+            #
+            # The text block is left-aligned to the right of the icon
+            # image.  If there is no image, the text is left-aliged
+            # with the left-hand border of the annotation
+            #
+            # display fields:
+            #   CN, DN, date, contact, reason, location
+            #
+            # Dict format:
+            #   appearance = dict(
+            #       background = Image with alpha / None,
+            #       icon = Image with alpha / None,
+            #       labels = bool,
+            #       display = list,
+            #       software = str,
+            #       outline = [R, G, B],
+            #       border = int,
+            #       )
+            sig = {}
+            for f in ("background", "icon", "labels", "border", "outline"):
+                if f in udct["signature_appearance"]:
+                    sig[f] = udct["signature_appearance"][f]
+
+            toggles = udct["signature_appearance"].get("display", [])
+            if(isinstance(toggles, str)):
+                sig["text"] = toggles
+            else:
+                for f in ("contact", "reason", "location", "contact", "signingdate"):
+                    if f in toggles:
+                        sig[f] = udct.get(f, "{} unknown".format(f))
+                if "date" in toggles:
+                    sig["date"] = udct["signingdate"]
+                if "CN" in toggles:
+                    sig["CN"] = cert.subject.get_attributes_for_oid(
+                        ObjectIdentifier("2.5.4.3")
+                    )[0].value
+                if "DN" in toggles:
+                    sig["DN"] = cert.subject.rfc4514_string()
+            annotation.simple_signature(sig)
+        else:
+            # Manual signature annotation creation
+            #
+            # Make your own appearance with an arbitrary number of
+            # images and fonts
+            if "manual_images" in udct:
+                for name, img in udct["manual_images"].items():
+                    annotation.add_image(img, name=name)
+            if "manual_fonts" in udct:
+                for name, path in udct["manual_fonts"].items():
+                    annotation.add_ttf_font(path, name=name)
+            annotation.add_default_font()
+            annotation.set_signature_appearance(*udct["signature_manual"])
+
+        pdfa = annotation.as_pdf_object(identity(), page=page0ref)
+        objapn = self._extend(pdfa["/AP"]["/N"])
+        objapnref = self._addObject(objapn)
+
+        objap = po.DictionaryObject()
+        objap[po.NameObject("/N")] = objapnref
+        obj13.update(
+            {
+                po.NameObject("/Rect"): po.ArrayObject(
+                    [
+                        po.FloatObject(x1),
+                        po.FloatObject(y1),
+                        po.FloatObject(x2),
+                        po.FloatObject(y2),
+                    ]
+                ),
+                po.NameObject("/AP"): objap,
+                # po.NameObject("/SM"): po.createStringObject("TabletPOSinline"),
+            }
+        )
+
+        page0 = page0ref.getObject()
+        if new_13:
+            annots = po.ArrayObject([obj13ref])
+            if "/Annots" in page0:
+                page0annots = page0["/Annots"]
+                if isinstance(page0annots, po.IndirectObject):
+                    annots.insert(0, page0annots)
+                elif isinstance(page0annots, po.ArrayObject):
+                    annots = page0annots
+                    annots.append(obj13ref)
+        else:
+            annots = page0["/Annots"]
+        page0.update({po.NameObject("/Annots"): annots})
+        self._objects[page0ref.idnum - 1] = page0
+
+    def makepdf(self, prev, udct, algomd, zeros, cert, othercerts, ocspurl, ocspissuer, **params):
         catalog = prev.trailer["/Root"]
         size = prev.trailer["/Size"]
         pages = catalog["/Pages"].getObject()
@@ -258,39 +420,63 @@ class SignedData(pdf.PdfFileWriter):
         while len(self._objects) < size - 1:
             self._objects.append(None)
 
-        ##        if params['mode'] == 'timestamp':
-        ##            # deal with extensions
-        ##            if '/Extensions' not in catalog:
-        ##                extensions = po.DictionaryObject()
-        ##            else:
-        ##                extensions = catalog['/Extensions']
-        ##
-        ##            if '/ESIC' not in extensions:
-        ##                extensions.update({
-        ##                    po.NameObject("/ESIC"): po.DictionaryObject({
-        ##                        po.NameObject('/BaseVersion'): po.NameObject('/1.7'),
-        ##                        po.NameObject('/ExtensionLevel'): po.NumberObject(1),
-        ##                        })
-        ##                    })
-        ##            else:
-        ##                esic = extensions['/ESIC']
-        ##                major, minor = esic['/BaseVersion'].lstrip('/').split('.')
-        ##                if int(major) < 1 or int(minor) < 7:
-        ##                    esic.update({
-        ##                        po.NameObject('/BaseVersion'): po.NameObject('/1.7'),
-        ##                        po.NameObject('/ExtensionLevel'): po.NumberObject(1),
-        ##                        })
-        ##            catalog.update({
-        ##                po.NameObject('/Extensions'): extensions
-        ##                })
+        if params['mode'] == 'timestamp':
+            # deal with extensions
+            if "/Extensions" not in catalog:
+                extensions = po.DictionaryObject()
+            else:
+                extensions = catalog["/Extensions"]
+
+            if "/ESIC" not in extensions:
+                extensions.update(
+                    {
+                        po.NameObject("/ESIC"): po.DictionaryObject(
+                            {
+                                po.NameObject("/BaseVersion"): po.NameObject("/1.7"),
+                                po.NameObject("/ExtensionLevel"): po.NumberObject(1),
+                            }
+                        )
+                    }
+                )
+                catalog.update({po.NameObject("/Extensions"): extensions})
+            else:
+                esic = extensions["/ESIC"]
+                major, minor = esic["/BaseVersion"].lstrip("/").split(".")
+                if int(major) < 1 or int(minor) < 7:
+                    esic.update(
+                        {
+                            po.NameObject("/BaseVersion"): po.NameObject("/1.7"),
+                            po.NameObject("/ExtensionLevel"): po.NumberObject(1),
+                        }
+                    )
+
+            SubFilter=po.NameObject("/ETSI.CAdES.detached")
+        else:
+            SubFilter=po.NameObject("/adbe.pkcs7.detached")
 
         # obj12 is the digital signature
         obj12, obj12ref = self._make_signature(
             Type=po.NameObject("/Sig"),
-            SubFilter=po.NameObject("/adbe.pkcs7.detached"),
+            SubFilter=SubFilter,
             Contents=UnencryptedBytes(zeros),
         )
 
+        obj12.update(
+            {
+                po.NameObject("/Prop_Build"): pdf.DictionaryObject(
+                    {
+                        po.NameObject("/App"): pdf.DictionaryObject(
+                            {
+                                po.NameObject("/Name"): po.NameObject(
+                                    "/" + udct.get("application", "endesive")
+                                )
+                            }
+                        ),
+                    }
+                ),
+            }
+        )
+
         if params["mode"] == "timestamp":
             # obj12 is a timestamp this time
             obj12.update(
@@ -300,18 +486,20 @@ class SignedData(pdf.PdfFileWriter):
                     po.NameObject("/V"): po.NumberObject(0),
                 }
             )
-        else:
-            obj12.update(
-                {
-                    po.NameObject("/Name"): po.createStringObject(udct["contact"]),
-                    po.NameObject("/Location"): po.createStringObject(udct["location"]),
-                    po.NameObject("/Reason"): po.createStringObject(udct["reason"]),
-                }
-            )
-            if params.get("use_signingdate"):
-                obj12.update(
-                    {po.NameObject("/M"): po.createStringObject(udct["signingdate"])}
-                )
+        d12 = {}
+        for k, v in (
+            ('/Name', 'name'),
+            ('/Contact', 'contact'),
+            ('/Location', 'location'),
+            ('/Reason', 'reason'),
+        ):
+            v = udct.get(v, None)
+            if v is not None:
+                d12[po.NameObject(k)] = po.createStringObject(v)
+        if 1 or params.get("use_signingdate"):
+            d12[po.NameObject("/M")] = po.createStringObject(udct["signingdate"])
+        if d12:
+            obj12.update(d12)
 
         # obj13 is a combined AcroForm Sig field with Widget annotation
         new_13 = True
@@ -353,161 +541,25 @@ class SignedData(pdf.PdfFileWriter):
 
         # add an annotation if there is a field to fill
         if box is not None:
-            from endesive.pdf.PyPDF2_annotate.annotations.signature import Signature
-            from endesive.pdf.PyPDF2_annotate.config.appearance import Appearance
-            from endesive.pdf.PyPDF2_annotate.config.location import Location
-            from endesive.pdf.PyPDF2_annotate.util.geometry import identity
-
-            x1, y1, x2, y2 = box
-            annotation = Signature(
-                Location(x1=x1, y1=y1, x2=x2, y2=y2, page=0), Appearance()
-            )
-            if "signature" in udct:
-                # Plain text signature with the default font
-                # text to render is contained in udct['signature']
-                # font parameters are in udct['signature']['text']
-                annotationtext = udct["signature"]
-                wrap_text = udct.get("text", {}).get("wraptext", True)
-                font_size = udct.get("text", {}).get("fontsize", 12)
-                text_align = udct.get("text", {}).get("textalign", "left")
-                line_spacing = udct.get("text", {}).get("linespacing", 1.2)
-
-                annotation.add_default_font()
-                annotation.set_signature_appearance(
-                    ["fill_colour", 0, 0, 0],
-                    ["font", "default", font_size],
-                    [
-                        "text_box",
-                        annotationtext,
-                        "default",
-                        0,
-                        0,
-                        x2 - x1,
-                        y2 - y1,
-                        font_size,
-                        wrap_text,
-                        text_align,
-                        "middle",
-                        line_spacing,
-                    ],
-                )
-            elif "signature_img" in udct:
-                # Simple image signature, stretches to fit the box
-                # image to render is contained in udct['signature_image']
-                annotation.add_image(udct["signature_img"], "Image")
-                annotation.set_signature_appearance(
-                    [
-                        "image",
-                        "Image",
-                        0,
-                        0,
-                        x2 - x1,
-                        y2 - y1,
-                        udct.get("signature_img_distort", True),
-                        udct.get("signature_img_centred", False),
-                    ]
-                )
-            elif "signature_appearance" in udct:
-                # Adobe-inspired signature with text and images
-                # Parameters are contained in udct['signature_appearance']
-                # If a field is included in the display list, that field
-                # will be contained in the annotation.
-                #
-                # Text and border are the colour specified by outline,
-                # and border is the the inset distance from the outer
-                # edge of the annotation.  The R G B values range between
-                # 0 and 1.
-                #
-                # Icon is an image to display above the background and
-                # border at the left-hand side of the anntoation.  If
-                # there is no text, it is centred.
-                #
-                # The text block is left-aligned to the right of the icon
-                # image.  If there is no image, the text is left-aliged
-                # with the left-hand border of the annotation
-                #
-                # display fields:
-                #   CN, DN, date, contact, reason, location
-                #
-                # Dict format:
-                #   appearance = dict(
-                #       background = Image with alpha / None,
-                #       icon = Image with alpha / None,
-                #       labels = bool,
-                #       display = list,
-                #       software = str,
-                #       outline = [R, G, B],
-                #       border = int,
-                #       )
-                sig = {}
-                for f in ("background", "icon", "labels", "border", "outline"):
-                    if f in udct["signature_appearance"]:
-                        sig[f] = udct["signature_appearance"][f]
-
-                toggles = udct["signature_appearance"].get("display", [])
-                for f in ("contact", "reason", "location", "contact", "signingdate"):
-                    if f in toggles:
-                        sig[f] = udct.get(f, "{} unknown".format(f))
-                if "date" in toggles:
-                    sig["date"] = udct["signingdate"]
-                if "CN" in toggles:
-                    from cryptography.x509 import ObjectIdentifier
-
-                    sig["CN"] = cert.subject.get_attributes_for_oid(
-                        ObjectIdentifier("2.5.4.3")
-                    )[0].value
-                if "DN" in toggles:
-                    sig["DN"] = cert.subject.rfc4514_string()
-                annotation.simple_signature(sig)
-            else:
-                # Manual signature annotation creation
-                #
-                # Make your own appearance with an arbitrary number of
-                # images and fonts
-                if "manual_images" in udct:
-                    for name, img in udct["manual_images"].items():
-                        annotation.add_image(img, name=name)
-                if "manual_fonts" in udct:
-                    for name, path in udct["manual_fonts"].items():
-                        annotation.add_ttf_font(path, name=name)
-                annotation.add_default_font()
-                annotation.set_signature_appearance(*udct["signature_manual"])
-
-            pdfa = annotation.as_pdf_object(identity(), page=page0ref)
-            objapn = self._extend(pdfa["/AP"]["/N"])
-            objapnref = self._addObject(objapn)
-
+            self.addAnnotation(cert, udct, box, page0ref, obj13, obj13ref, new_13)
+        else:
+            # invisible signature
             objap = po.DictionaryObject()
-            objap[po.NameObject("/N")] = objapnref
-            obj13.update(
-                {
-                    po.NameObject("/Rect"): po.ArrayObject(
-                        [
-                            po.FloatObject(x1),
-                            po.FloatObject(y1),
-                            po.FloatObject(x2),
-                            po.FloatObject(y2),
-                        ]
-                    ),
-                    po.NameObject("/AP"): objap,
-                    # po.NameObject("/SM"): po.createStringObject("TabletPOSinline"),
-                }
-            )
-
-            page0 = page0ref.getObject()
-            if new_13:
-                annots = po.ArrayObject([obj13ref])
-                if "/Annots" in page0:
-                    page0annots = page0["/Annots"]
-                    if isinstance(page0annots, po.IndirectObject):
-                        annots.insert(0, page0annots)
-                    elif isinstance(page0annots, po.ArrayObject):
-                        annots = page0annots
-                        annots.append(obj13ref)
-            else:
-                annots = page0["/Annots"]
-            page0.update({po.NameObject("/Annots"): annots})
-            self._objects[page0ref.idnum - 1] = page0
+            objapref = self._addObject(objap)
+            objform = po.DictionaryObject({
+                po.NameObject("/Length"): po.NumberObject(0),
+                po.NameObject("/Type"): po.NameObject("/XObject"),
+                po.NameObject("/Subtype"): po.NameObject("/Form"),
+                po.NameObject("/Bbox"): po.ArrayObject([
+                    po.FloatObject(0.0),
+                    po.FloatObject(0.0),
+                    po.FloatObject(0.0),
+                    po.FloatObject(0.0),
+                ]),
+            })
+            objformref = self._addObject(objform)
+            objap.update({po.NameObject("/N"): objformref})
+            obj13.update({po.NameObject("/AP"): objapref})
 
         if udct.get("sigandcertify", False) and "/Perms" not in catalog:
             obj10 = po.DictionaryObject()
@@ -553,7 +605,7 @@ class SignedData(pdf.PdfFileWriter):
                 for i in range(1, len(checklist) + 1):
                     suffix = "_{}".format(i)
                     if suffix in checklist:
-                        next
+                        continue
 
                     new_name = "{}{}".format(name_base, suffix)
                     obj13.update({po.NameObject("/T"): EncodedString(new_name)})
@@ -590,6 +642,43 @@ class SignedData(pdf.PdfFileWriter):
         if "/Metadata" in catalog:
             catalog[po.NameObject("/Metadata")] = catalog.raw_get("/Metadata")
 
+        if udct.get("ltv", False):
+            dss = po.DictionaryObject()
+            vri = po.DictionaryObject()
+            certs = po.ArrayObject()
+            ocsps = po.ArrayObject()
+            crls = po.ArrayObject()
+            dss.update({
+                po.NameObject("/Type"): po.NameObject("/DSS"),
+                po.NameObject("/VRI"): vri,
+                po.NameObject("/Certs"): certs,
+                po.NameObject("/OCSPs"): ocsps,
+                po.NameObject("/CRLs"): crls,
+            })
+            catalog[po.NameObject("/DSS")] = self._addObject(dss)
+
+            if ocspurl is None:
+                ocspurl = signer.extract_ocsp_url_from_cert(cert)
+            if ocspurl is not None:
+                if ocspissuer is None:
+                    for othercert in othercerts:
+                        if othercert.subject == cert.issuer:
+                            ocspissuer = othercert
+                            break
+                certissuer = ocspissuer
+                ocspresp = signer.fetch_ocsp_response(cert, certissuer, ocspurl)
+            if ocspresp is not None:
+                obj = po.StreamObject()
+                obj._data = ocspresp
+                ocsps.append(self._addObject(obj))
+            obj = po.StreamObject()
+            obj._data = cert.public_bytes(serialization.Encoding.DER)
+            certs.append(self._addObject(obj))
+            for cert in othercerts:
+                obj = po.StreamObject()
+                obj._data = cert.public_bytes(serialization.Encoding.DER)
+                certs.append(self._addObject(obj))
+
         x_root = prev.trailer.raw_get("/Root")
         self._objects[x_root.idnum - 1] = catalog
         self.x_root = po.IndirectObject(x_root.idnum, 0, self)
@@ -604,12 +693,13 @@ class SignedData(pdf.PdfFileWriter):
         othercerts,
         algomd,
         hsm,
-        timestampurl,
+        timestampurl=None,
         timestampcredentials=None,
         timestamp_req_options=None,
         mode="sign",
         ocspurl=None,
         ocspissuer=None,
+        use_signingdate=True,
     ):
         startdata = len(datau)
 
@@ -637,7 +727,7 @@ class SignedData(pdf.PdfFileWriter):
             algomd = obj["/DigestMethod"][1:].lower()
 
         # produce smaller signatures, but must be signed twice
-        aligned = udct.get("aligned", 0)
+        aligned = udct.get("aligned", 4096 if isinstance(key, ec.EllipticCurvePrivateKey) else 0)
         if aligned:
             zeros = b"00" * aligned
         else:
@@ -652,16 +742,18 @@ class SignedData(pdf.PdfFileWriter):
                     prehashed=md,
                 )[0]["values"][0].dump()
             else:
+                attrs = udct.get("attrs", True)
+                pss = udct.get("pss", False)
                 contents = signer.sign(
                     None,
                     key,
                     cert,
                     othercerts,
                     algomd,
-                    True,
+                    attrs,
                     md,
                     hsm,
-                    False,
+                    pss,
                     timestampurl,
                     timestampcredentials,
                     timestamp_req_options,
@@ -670,11 +762,11 @@ class SignedData(pdf.PdfFileWriter):
                 )
             zeros = contents.hex().encode("utf-8")
 
-        params = {"mode": mode}
+        params = {"mode": mode, "use_signingdate": use_signingdate}
         if not timestampurl:
             params["use_signingdate"] = True
 
-        self.makepdf(prev, udct, algomd, zeros, cert, **params)
+        self.makepdf(prev, udct, algomd, zeros, cert, othercerts, ocspurl, ocspissuer, **params)
 
         # if document was encrypted, encrypt this version too
         if prev.isEncrypted:
@@ -685,15 +777,14 @@ class SignedData(pdf.PdfFileWriter):
         # ID[0] is used in password protection, must be unchanged
         ID = prev.trailer.get("/ID", None)
         if ID is None:
-            ID = hashlib.md5(repr(time.time()).encode()).digest()
+            ID = udct.get("id") or hashlib.md5(repr(time.time()).encode()).digest()
         else:
             ID = ID.getObject()[0].original_bytes
+        newID = udct.get("newid", repr(random.random()))
         self._ID = po.ArrayObject(
             [
                 po.ByteStringObject(ID),
-                po.ByteStringObject(
-                    hashlib.md5(repr(random.random()).encode()).digest()
-                ),
+                po.ByteStringObject(hashlib.md5(newID.encode()).digest()),
             ]
         )
 
@@ -735,16 +826,18 @@ class SignedData(pdf.PdfFileWriter):
                 prehashed=md,
             )[0]["values"][0].dump()
         else:
+            attrs = udct.get("attrs", True)
+            pss = udct.get("pss", False)
             contents = signer.sign(
                 None,
                 key,
                 cert,
                 othercerts,
                 algomd,
-                True,
+                attrs,
                 md,
                 hsm,
-                False,
+                pss,
                 timestampurl,
                 timestampcredentials,
                 timestamp_req_options,
@@ -836,7 +929,7 @@ def sign(
     timestampcredentials=None,
     timestamp_req_options=None,
     ocspurl=None,
-    ocspissuer=None
+    ocspissuer=None,
 ):
     """
     parameters:
@@ -895,6 +988,8 @@ def sign(
             password: string            required if the document is password protected, signing it also requires that password
             text: dict                  text attributes
                                             wraptext=True, fontsize:12, textalign:'left', linespacing:1.2
+            application: string         optional application name in advanced signature properties dialog in Acrobat Reader
+            pss: boolen                 optional choosen signature scheme RSASSA-PSS (true) vs. PKCS1-v1_5 (false/default)
         key: cryptography.hazmat.backends.openssl.rsa._RSAPrivateKey - private key used to sign the document
         cert: cryptography.x509.Certificate - certificate associated with the key
         othercerts: list of cryptography.x509.Certificate to be saved with the signed document,
@@ -907,6 +1002,7 @@ def sign(
         timestamp_req_options: Dict to set options to the POST http call against the timestamp server. Default: None
         ocsppurl: ocsp server URL or None
         ocspissuer: certificate of issuer or None
+        ltv: boolean enable LTV signature
 
     returns: bytes ready for writing after unsigned pdf document containing its electronic signature
 
@@ -975,6 +1071,5 @@ def sign(
         timestamp_req_options,
         "sign",
         ocspurl,
-        ocspissuer
+        ocspissuer,
     )
-
