@@ -5,16 +5,15 @@ import sys
 import types
 import hashlib
 import time
+import requests
 from base64 import b64encode
 from datetime import datetime
-
-import requests
-import pytz
 from asn1crypto import cms, algos, core, keys, pem, tsp, x509, ocsp, util
-from oscrypto import asymmetric
-from cryptography.hazmat import backends
 from cryptography.hazmat.primitives import hashes, serialization
-from cryptography.hazmat.primitives.asymmetric import padding, utils
+from cryptography.hazmat.primitives.asymmetric import padding, utils, ec
+from cryptography.hazmat import backends
+from cryptography import x509 as cryptography_x509
+from cryptography.x509 import ocsp as cryptography_ocsp
 
 
 def cert2asn(cert, cert_bytes=True):
@@ -29,6 +28,58 @@ def cert2asn(cert, cert_bytes=True):
     return x509.Certificate.load(cert_bytes)
 
 
+def extract_ocsp_url_from_cert(cert):
+    """Extract OCSP URL from certificate's Authority Information Access extension"""
+    if hasattr(cert, 'public_bytes'):
+        crypto_cert = cert
+    else:
+        if hasattr(cert, 'dump'):
+            cert_bytes = cert.dump()
+        else:
+            cert_bytes = cert
+        crypto_cert = cryptography_x509.load_der_x509_certificate(
+            cert_bytes, backends.default_backend())
+
+    try:
+        aia = crypto_cert.extensions.get_extension_for_oid(
+            cryptography_x509.oid.ExtensionOID.AUTHORITY_INFORMATION_ACCESS)
+        for access_description in aia.value:
+            if access_description.access_method == cryptography_x509.oid.AuthorityInformationAccessOID.OCSP:
+                return access_description.access_location.value
+    except cryptography_x509.ExtensionNotFound:
+        return None
+    return None
+
+
+def fetch_ocsp_response(cert, issuer, url):
+    if hasattr(cert, 'dump'):
+        cert_bytes = cert.dump()
+        cert = cryptography_x509.load_der_x509_certificate(
+            cert_bytes, backends.default_backend())
+
+    if hasattr(issuer, 'dump'):
+        issuer_bytes = issuer.dump()
+        issuer = cryptography_x509.load_der_x509_certificate(
+            issuer_bytes, backends.default_backend())
+
+    builder = cryptography_ocsp.OCSPRequestBuilder()
+    builder = builder.add_certificate(cert, issuer, hashes.SHA1())
+    req = builder.build()
+    data = req.public_bytes(serialization.Encoding.DER)
+
+    try:
+        response = requests.post(
+            url,
+            headers={"Content-Type": "application/ocsp-request"},
+            data=data,
+        )
+        if response.status_code != 200:
+            return None
+        return response.content
+    except requests.exceptions.ConnectionError:
+        return None
+
+
 def timestamp(unhashed, hashalgo, url, credentials, req_options, prehashed=None):
     if prehashed:
         hashed_value = prehashed
@@ -151,24 +202,28 @@ def sign(
             {"algorithm": "rsassa_pkcs1v15"}
         )
     else:
+        md = getattr(hashes, hashalgo.upper())
         if isinstance(key, keys.PrivateKeyInfo):
-            salt_length = key.byte_size - hashes.SHA512.digest_size - 2
-            salt_length = hashes.SHA512.digest_size
+            salt_length = key.byte_size - md.digest_size - 2
+            salt_length = md.digest_size
         else:
-            salt_length = padding.calculate_max_pss_salt_length(key, hashes.SHA512)
+            if key is None:
+                salt_length = md.digest_size
+            else:
+                salt_length = padding.calculate_max_pss_salt_length(key, md)
         signer["signature_algorithm"] = algos.SignedDigestAlgorithm(
             {
                 "algorithm": "rsassa_pss",
                 "parameters": algos.RSASSAPSSParams(
                     {
                         "hash_algorithm": algos.DigestAlgorithm(
-                            {"algorithm": "sha512"}
+                            {"algorithm": hashalgo.lower()}
                         ),
                         "mask_gen_algorithm": algos.MaskGenAlgorithm(
                             {
                                 "algorithm": algos.MaskGenAlgorithmId("mgf1"),
                                 "parameters": {
-                                    "algorithm": algos.DigestAlgorithmId("sha512"),
+                                    "algorithm": algos.DigestAlgorithmId(hashalgo.lower()),
                                 },
                             }
                         ),
@@ -290,37 +345,22 @@ def sign(
         ],
     }
     if ocspurl and ocspissuer:
-        from cryptography.hazmat.backends.openssl.backend import backend
-        from cryptography.x509 import ocsp as cocsp
-        from cryptography import x509 as cx509
-
-        ocspuser = cert.dump()
-        ocspuser = cx509.load_der_x509_certificate(ocspuser, backend=backend)
-
-        builder = cocsp.OCSPRequestBuilder()
-        builder = builder.add_certificate(ocspuser, ocspissuer, hashes.SHA1())
-        req = builder.build()
-        data = req.public_bytes(serialization.Encoding.DER)
-
-        response = requests.post(
-            ocspurl,
-            headers={"Content-Type": "application/ocsp-request"},
-            data=data,
-        )
-        data = ocsp.OCSPResponse.load(response.content)
-        other = cms.RevocationInfoChoice(
-            {
-                "other": cms.OtherRevocationInfoFormat(
-                    {
-                        "other_rev_info_format": cms.OtherRevInfoFormatId(
-                            "ocsp_response"
-                        ),
-                        "other_rev_info": data,
-                    }
-                )
-            }
-        )
-        config["crls"] = cms.RevocationInfoChoices([other])
+        ocsp_response = fetch_ocsp_response(cert, ocspissuer, ocspurl)
+        if ocsp_response:
+            ocsp_response = ocsp.OCSPResponse.load(ocsp_response)
+            other = cms.RevocationInfoChoice(
+                {
+                    "other": cms.OtherRevocationInfoFormat(
+                        {
+                            "other_rev_info_format": cms.OtherRevInfoFormatId(
+                                "ocsp_response"
+                            ),
+                            "other_rev_info": ocsp_response,
+                        }
+                    )
+                }
+            )
+            config["crls"] = cms.RevocationInfoChoices([other])
 
     datas = cms.ContentInfo(
         {
@@ -335,28 +375,26 @@ def sign(
         tosign = datau
     if hsm is not None:
         signed_value_signature = hsm.sign(keyid, tosign, hashalgo)
-    elif isinstance(key, keys.PrivateKeyInfo):
-        key = asymmetric.load_private_key(key)
-        if pss:
-            signed_value_signature = asymmetric.rsa_pss_sign(key, tosign, "sha512")
-        else:
-            signed_value_signature = asymmetric.rsa_pkcs1v15_sign(
-                key, tosign, hashalgo.lower()
-            )
     else:
         if pss:
-            hasher = hashes.Hash(hashes.SHA512(), backend=backends.default_backend())
+            md = getattr(hashes, hashalgo.upper())
+            hasher = hashes.Hash(md(), backend=backends.default_backend())
             hasher.update(tosign)
             digest = hasher.finalize()
             signed_value_signature = key.sign(
                 digest,
-                padding.PSS(mgf=padding.MGF1(hashes.SHA512()), salt_length=salt_length),
-                utils.Prehashed(hashes.SHA512()),
+                padding.PSS(mgf=padding.MGF1(md()), salt_length=salt_length),
+                utils.Prehashed(md()),
             )
         else:
-            signed_value_signature = key.sign(
-                tosign, padding.PKCS1v15(), getattr(hashes, hashalgo.upper())()
-            )
+            if isinstance(key, ec.EllipticCurvePrivateKey):
+                signed_value_signature = key.sign(
+                    tosign, ec.ECDSA(getattr(hashes, hashalgo.upper())())
+                )
+            else:
+                signed_value_signature = key.sign(
+                    tosign, padding.PKCS1v15(), getattr(hashes, hashalgo.upper())()
+                )
 
     if timestampurl is not None:
         datas["content"]["signer_infos"][0]["unsigned_attrs"] = timestamp(
