@@ -24,11 +24,11 @@ def main():
         # "sigbutton": True,
         "sigfield": "Signature1",
         "auto_sigfield": True,
-        # "sigandcertify": True,
+        "sigandcertify": True,
         "signaturebox": (470, 840, 570, 640),
-        "signature": "Dokument podpisany cyfrowo ąćęłńóśżź",
+        "signature": "Podpis 1",
         # "signature_img": "signature_test.png",
-        "contact": "contact:mak@trisoft.com.pl",
+        "contact": "demo1@demot.com.pl",
         "location": "Szczecin",
         "signingdate": date,
         "reason": "Dokument podpisany cyfrowo aą cć eę lł nń oó sś zż zź",
